/*
PRICE was written by Graham Ruby.
This file is part of PRICE.

PRICE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

PRICE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PRICE.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef ECOFILTER_CPP
#define ECOFILTER_CPP

#include "EcoFilter.h"
#include "AssemblyException.h"
#include "Alignment.h"
#include "ScoredSeqCollectionBwt.h"
#include "MatchSeqTest.h"
#include "AssemblyJobPerfect.h"

EcoFilter::~EcoFilter(){}


EcoFilterNull::EcoFilterNull(){}
EcoFilterNull::~EcoFilterNull(){}
EcoFilterNull* EcoFilterNull::copy(){ return new EcoFilterNull(); }
void EcoFilterNull::filterSeqs(set<ScoredSeq*>* fullSet,
			       set<ScoredSeq*>* retainedSet,
			       set<ScoredSeq*>* removedSet,
			       int cycleNum){
  retainedSet->insert(fullSet->begin(), fullSet->end());
}


EcoFilterMinCount::EcoFilterMinCount(){}
EcoFilterMinCount::EcoFilterMinCount(float minCount) : _minCount(minCount){}
EcoFilterMinCount::~EcoFilterMinCount(){}
EcoFilterMinCount* EcoFilterMinCount::copy(){ return new EcoFilterMinCount(_minCount); }
void EcoFilterMinCount::filterSeqs(set<ScoredSeq*>* fullSet,
				   set<ScoredSeq*>* retainedSet,
				   set<ScoredSeq*>* removedSet,
				   int cycleNum){
  for (set<ScoredSeq*>::iterator seqIt = fullSet->begin(); seqIt != fullSet->end(); ++seqIt){
    float maxScore = 0;
    ScoredSeq* seq = (*seqIt);
    for (int n = 0; n < seq->size(); ++n){
      float nucScore = seq->scoreAtPosition(n,'+');
      if (nucScore > maxScore){ maxScore = nucScore; }
      if (n < seq->size() - 1){
	float linkage = seq->linkAfterPosition(n,'+');
	if (linkage > maxScore){ maxScore = linkage; }
      }
    }
    if (maxScore < _minCount){ removedSet->insert(seq); }
    else { retainedSet->insert(seq); }
  }
}
EcoFilterMinCount* EcoFilterMinCount::defaultFilter(){
  return new EcoFilterMinCount(1.6);
}




EcoFilterMinLength::EcoFilterMinLength(){}
EcoFilterMinLength::EcoFilterMinLength(long minLength, int skipCycles) : _minLength(minLength), _skipCycles(skipCycles){}
EcoFilterMinLength::~EcoFilterMinLength(){}
EcoFilterMinLength* EcoFilterMinLength::copy(){ return new EcoFilterMinLength(_minLength,_skipCycles); }
void EcoFilterMinLength::filterSeqs(set<ScoredSeq*>* fullSet,
				    set<ScoredSeq*>* retainedSet,
				    set<ScoredSeq*>* removedSet,
				    int cycleNum){
  if (cycleNum < _skipCycles){
    retainedSet->insert(fullSet->begin(), fullSet->end());
  } else {
    for (set<ScoredSeq*>::iterator seqIt = fullSet->begin(); seqIt != fullSet->end(); ++seqIt){
      if ( (*seqIt)->size() < _minLength ){ removedSet->insert( *seqIt ); }
      else { retainedSet->insert( *seqIt ); }
    }
  }
}





EcoFilterInitialContigMatch::EcoFilterInitialContigMatch(){}
EcoFilterInitialContigMatch::EcoFilterInitialContigMatch(float minFractId, int cyclesToSkip, bool fullFile) : 
  _minFractId(minFractId),
  _cyclesToSkip(cyclesToSkip),
  _numFilterCycles(1),
  _numSkipCycles(0),
  _fullFile(fullFile)
{}
EcoFilterInitialContigMatch::EcoFilterInitialContigMatch(float minFractId, int cyclesToSkip,
							 int numFilterCycles, int numSkipCycles, bool fullFile) :
  _minFractId(minFractId),
  _cyclesToSkip(cyclesToSkip),
  _numFilterCycles(numFilterCycles),
  _numSkipCycles(numSkipCycles),
  _fullFile(fullFile)
{}
EcoFilterInitialContigMatch::~EcoFilterInitialContigMatch(){}
EcoFilterInitialContigMatch* EcoFilterInitialContigMatch::copy(){ 
  EcoFilterInitialContigMatch* copy = new EcoFilterInitialContigMatch(_minFractId,_cyclesToSkip,
								      _numFilterCycles,_numSkipCycles,_fullFile);
  copy->addFiles(&_initialContigFiles);
  return copy;
}

void EcoFilterInitialContigMatch::addFiles(set<ParameterizedInitialFile*>* initialContigFiles){
  _initialContigFiles.insert(initialContigFiles->begin(), initialContigFiles->end());
}


void EcoFilterInitialContigMatch::filterSeqs(set<ScoredSeq*>* fullSet,
					     set<ScoredSeq*>* retainedSet,
					     set<ScoredSeq*>* removedSet,
					     int cycleNum){
  // too early in the assembly, so nothing happens
  if (cycleNum < _cyclesToSkip or
      (cycleNum - _cyclesToSkip) % (_numFilterCycles + _numSkipCycles) >= _numFilterCycles ){
    retainedSet->insert(fullSet->begin(), fullSet->end());

  } else {
    // get the input contigs from all of the prior cycles AND the current cycle
    set<ScoredSeq*> inputContigs;
    for (set<ParameterizedInitialFile*>::iterator it = _initialContigFiles.begin(); it != _initialContigFiles.end(); ++it){
      if (_fullFile){
	for (int n = 0; n <= (*it)->totalCycles(); ++n){ (*it)->getContigs(&inputContigs, n); }
      } else {
	for (int n = 0; n <= cycleNum; ++n){ (*it)->getContigs(&inputContigs, n); }
      }
    }

    // remove redundant sequences
    AssemblyJob* aj = new AssemblyJobPerfect(&inputContigs, AssemblyJob::doubleStranded);
    aj->runJob(AssemblyJob::notThreaded);
    // delete the discarded seqs, then collect the legit ones
    set<ScoredSeq*> ajOutput;
    aj->discardedInputSeqs( &ajOutput );
    for (set<ScoredSeq*>::iterator it = ajOutput.begin(); it != ajOutput.end(); ++it){ (*it)->deepDelete(); }
    ajOutput.clear();
    aj->allAssembledSeqs( &ajOutput );
    delete aj;

    // make a collection of the existing contigs
    // WAS UNGAPPED
    //ScoredSeqCollectionBwt* contigSet = new ScoredSeqCollectionBwt(fullSet, _minFractId);
    // NOW GAPPED
    AlignmentScoreMatrix* asMatrix = AlignmentScoreMatrix::getDefault();
    DynamicProgrammingAligner* dpa = new DynamicProgrammingAligner(_minFractId, 0, asMatrix);
    delete asMatrix;
    ScoredSeqCollectionBwt* contigSet = new ScoredSeqCollectionBwt(fullSet, dpa);
    delete dpa;

    set<ScoredSeq*>::iterator seqIt = ajOutput.begin();
    // this can end if all of the contigs have been validated
    while (seqIt != ajOutput.end() and retainedSet->size() < fullSet->size()){
      ScoredSeq* initSeq = *seqIt;
      ++seqIt;
      set<Alignment*> matches;
      MatchSeqTest* matchTest = new MatchSeqTestNotInSet(retainedSet);
      contigSet->getMatches(&matches, initSeq, matchTest, initSeq->size(), ScoredSeqCollectionBwt::softMinOvl);
      // the hit contigs will now be retained
      for (set<Alignment*>::iterator alIt = matches.begin(); alIt != matches.end(); ++alIt){
	retainedSet->insert( (*alIt)->seqB() );
	delete *alIt;
      }
      delete matchTest;
    }

    // clean up
    for (set<ScoredSeq*>::iterator it = ajOutput.begin(); it != ajOutput.end(); ++it){ (*it)->deepDelete(); }
    delete contigSet;
    for (set<ScoredSeq*>::iterator it = fullSet->begin(); it != fullSet->end(); ++it){
      if (retainedSet->count(*it) == 0){ removedSet->insert(*it); }
    }
  }
}



EcoFilterMulti::EcoFilterMulti(){}
EcoFilterMulti::EcoFilterMulti(EcoFilter** testArray, int numTests) : _numTests(numTests){
  _testArray = new EcoFilter*[ _numTests ];
  for (int n = 0; n < _numTests; ++n){ _testArray[n] = testArray[n]->copy(); }
}
EcoFilterMulti::~EcoFilterMulti(){
  for (int testNum = 0; testNum < _numTests; ++testNum){ delete _testArray[testNum]; }
  delete [] _testArray;
}
EcoFilterMulti* EcoFilterMulti::copy(){ return new EcoFilterMulti(_testArray,_numTests); }
void EcoFilterMulti::filterSeqs(set<ScoredSeq*>* fullSet,
				set<ScoredSeq*>* retainedSet,
				set<ScoredSeq*>* removedSet,
				int cycleNum){
  set<ScoredSeq*> runningSet;
  runningSet.insert(fullSet->begin(), fullSet->end());
  for (int testNum = 0; testNum < _numTests; ++testNum){
    set<ScoredSeq*> tempSuccess;
    _testArray[testNum]->filterSeqs(&runningSet, &tempSuccess, removedSet, cycleNum);
    runningSet.clear();
    runningSet.insert(tempSuccess.begin(), tempSuccess.end());
  }
  retainedSet->insert(runningSet.begin(), runningSet.end());
}


#endif

